(******************************************************************)
(* Copyright 1997, Microchip Systems / Carl Bunton                *)
(* Email: Twojags@cris.com                                        *)
(* Web-page: http://www.concentric.net/~twojags                   *)
(*                                                                *)
(* This code is not for redistribution in whole or in part.  It   *)
(* may be used in compiled program format only.                   *)
(*                                                                *)
(* This program was written in Delphi 2 because version 2         *)
(* compiles a much smaller executable using the windows api.  It  *)
(* should be fully compatible with Delphi 3, but will produce a   *)
(* noticable increase of size in the final compiled program.      *)
(******************************************************************)
UNIT SFXgbls;

{$A-}

INTERFACE


USES
	Messages,
	Windows;


TYPE
  TLocal = RECORD
    	SignAtr       	: LONGINT;
    	VerNum        	: WORD;   (* VersionMadeBy *)
     BitFlag        : WORD;
     CompressType   : WORD;
     FileDate       : LONGINT;
     crc32          : LONGINT;
     PackedSize     : LONGINT;
     UnpackedSize   : LONGINT;
     FilenameLen    : WORD;
     ExtraFieldLen  : WORD;
  END;



	 TYPE
     pCRC32Table		= ^TCRC32;
     TCRC32 			= ARRAY[0..256] OF LONGINT;
   VAR
     CRC32Table 	: pCRC32Table;
     Crc32Val    : LONGINT;   { Running CRC (32 bit) value }


VAR
	key : ARRAY[0..2] OF LONGINT;


CONST
	WSIZE         = 32768;	 	{ window size--must be a power OF two, and at least 32k}

  (* String Table constants *)
  BHSFX 		  = 1;
  ZIPSFX		  = 2;
  ARJSFX		  = 3;
  LHASFX		  = 4;

	RAND_HEAD_LEN = 12;
	CRC_MASK      = $FFFFFFFF;
	HEADER_SIGNATURE 	= $04034B50;


VAR
  Bytes_To_Go : LONGINT;
  InFile		: THandle;
  OutFile		: THandle;
	Header		: TLocal;



	FUNCTION  ReadHeader: BOOLEAN;
	FUNCTION  GetArchiveOffset: LONGINT;
	PROCEDURE ProcessArchive ( DlgWin: hWnd; FillListBox: BOOLEAN  );
	FUNCTION  decrypt_pw(Encrypt_Head: PCHAR; EncHead_len: BYTE;
       BitFlag: WORD; CRC, FileDate: LONGINT; password: STRING): BOOLEAN;
	FUNCTION  decrypt_byte: INTEGER;
	PROCEDURE seedk(passwd: STRING);
	PROCEDURE UDK(c: byte);
	PROCEDURE Make_CRC32Table;
	FUNCTION  UpdC32(Octet: byte; Crc: LongInt) : LongInt;



IMPLEMENTATION



USES
	SFXinflt,
	Dialog,
	SFXmisc;



(*--------------------------------------------------------------------------*)
FUNCTION  ReadHeader: BOOLEAN;

VAR
	BytesRead: INTEGER;

BEGIN

  TRY
     Header.SignAtr := 0;
     ReadFile( InFile, Header, SIZEOF( Header ), BytesRead, NIL );

		RESULT := ( Header.SignAtr = HEADER_SIGNATURE ) OR
     	( Header.SignAtr = $50000000 );

     IF RESULT THEN
     	Bytes_To_Go := Header.PackedSize
     ELSE
     	Bytes_To_Go := 0;

  EXCEPT
     RESULT := FALSE;
  END;

END (* ReadHeader *);
(*--------------------------------------------------------------------------*)
FUNCTION  GetArchiveOffset: LONGINT;

VAR
  pFile       : PCHAR;

BEGIN

 	GetMem( pFile, 255 );

	TRY

     LoadString( hInstance, ZIPSFX, pFile, 255 );

  	RESULT := StrToInt( PCharToStr( pFile, StrLen( pFile ) ) );

   	IF GetFileSize( InFile, NIL ) = RESULT  THEN
     	RESULT := 0;

  FINALLY
  	Dispose( pFile );
  END;

END (* GetArchiveOffset *);
(*--------------------------------------------------------------------------*)
PROCEDURE ProcessArchive ( DlgWin: hWnd; FillListBox: BOOLEAN  );

VAR
	s			  : PCHAR;
	FN			  : PCHAR;
  IsPassword : BOOLEAN;
  EncryptHDR : PCHAR;
  Directory  : STRING;
  Filename	  : STRING;
  ExtractFile: STRING;
  hWndList	  : hWnd;
  i			  : BYTE;
  SavePos	  : LONGINT;
  BytesRead  : INTEGER;
	Msg 		  : TMsg;

LABEL
	ByPass;

BEGIN

	FSeek( StartOfFile, FILE_BEGIN  );
  IF FilePos = -1 THEN
  BEGIN
  	MessageBox( 0, 'Error finding attached archive.', 'Archive error..', mb_OK  );
     Exit;
  END;
  									(* Get handle of ArchiveFileListbox *)

  hWndList := GetDlgItem( DlgWin, CM_FILELIST  );
  GetMem ( FN, 256 + 2  );

  TRY

     								(* Start search at beginning of listbox *)

     Index := -1;
  	HasStoredPaths := FALSE;

										(* Clear the listbox *)

     IF FillListBox THEN
    		SendMessage ( hWndList, LB_RESETCONTENT, 0, 0 );

        WHILE ReadHeader do
        BEGIN

           SavePos := FilePos;

           ReadFile( InFile, FN^, Header.FilenameLen, BytesRead, NIL  );
           Filename := PCharToStr( FN, Header.FilenameLen  );

           IF Filename = '' THEN
           BEGIN
              MessageBox( 0, 'Error retreiving filename.', 'Archive error..', mb_OK  );
              EXIT;
           END;

           IF NOT HasStoredPaths THEN
              IF ( Pos( '\', Filename  ) > 0  ) OR
                    ( Pos( '/', Filename  ) > 0  ) THEN
                 HasStoredPaths := TRUE;


                                (* If Use Stored Dirs checkbox not checked, cut dir *)

           IF SendMessage( GetDlgItem( DlgWin, CM_STOREDPATH  ), BM_GETCHECK, 0, 0  ) <> 1 THEN
              Filename := ExtractFilename( Filename  );


           IF FillListBox THEN

                                (* Add each string to the listbox *)

              SendMessage ( hWndList, LB_ADDSTRING, 0, LONGINT( PCHAR( Filename  )  )  )

           ELSE BEGIN

              IF IsSelected( hWndList, PCHAR( Filename )  ) THEN
              BEGIN

                                (* Default *)

                 IsPassword := FALSE;

                 IF ( Header.BitFlag AND 1  ) = 1 THEN
                 BEGIN

                    DEC( Bytes_To_Go, RAND_HEAD_LEN  );

                    TRY

                       GetMem ( EncryptHDR, RAND_HEAD_LEN * 2  );
                       ReadFile( InFile, EncryptHDR^, RAND_HEAD_LEN, BytesRead, NIL  );

                                (* make a working copy of encrypted header in upper half of buffer *)

                       Move( EncryptHDR[0], EncryptHDR[RAND_HEAD_LEN], RAND_HEAD_LEN );

                       TRY

                          IF PW <> '' then
                             IsPassword := decrypt_pw( EncryptHDR,
                                                        RAND_HEAD_LEN,
                                                        Header.BitFlag,
                                                        Header.CRC32,
                                                        Header.FileDate,
                                                        PW
                                                        );

                          IF NOT IsPassword THEN
                             FOR i := 0 TO 2 DO
                             BEGIN

                                DialogBox( hInstance, 'PWD', 0, @PwdProc  );
                                PW := PCharToStr( Password, PWLen  );
                                IsPassword := decrypt_pw( EncryptHDR,
                                                           RAND_HEAD_LEN,
                                                           Header.BitFlag,
                                                           Header.CRC32,
                                                           Header.FileDate,
                                                           PW
                                                           );

                                IF IsPassword THEN BREAK;

                             END;

                       FINALLY

                          FreeMem ( EncryptHDR  );

                       END;

                    EXCEPT
                    END;

                 END ELSE
                                (* Not pw protected.. set value to extract *)
                    IsPassword := TRUE;



                 IF IsPassword THEN
                 BEGIN

                    IF SendMessage( GetDlgItem( DlgWin, CM_STOREDPATH  ), BM_GETCHECK, 0, 0  ) <> 1 THEN
                       ExtractFile := PCharToStr( AppendDirTail( ExtPath  ), StrLen( AppendDirTail( ExtPath  )  )  )
                          + Filename
                    ELSE
                       ExtractFile := PCharToStr( RemoveDirTail( ExtPath  ), StrLen( AppendDirTail( ExtPath  )  )  )
                          + Filename;

                    IF FileExists( PCHAR( ExtractFile  )  ) THEN
                    BEGIN

                       CASE OverWriteMode OF

                          //0:              (* Overwrite *)

                          1: BEGIN          (* Skip *)
                                SendMessage ( hWndList, LB_SETSEL, 0, Index  );
                                GOTO ByPass;
                             END;

                          2: BEGIN	  			(* Confirm *)

                                CurrentFile := ExtractFilename( ExtractFile  );
                                DialogBox( hInstance, 'FileExist', 0, @FileExistsProc  );

                                IF NOT OverWriteFile THEN
                                BEGIN
                                   SendMessage ( hWndList, LB_SETSEL, 0, Index  );
                                   GOTO ByPass;
                                END;

                             END;

                       END;

                    END ELSE BEGIN

                       Directory := ExtractFilePath( ExtractFile  );
                       ForceDirectories( Directory  );	(* removes trailing dir char *)

                       IF NOT FileExists( PCHAR( Directory  )  ) THEN
                       BEGIN
                       	MessageBox( 0, PCHAR( Directory  ), 'Error Changing Dir', mb_OK  );
                          BREAK;
                       END;

                    END;

                    OutFile := CreateFile( PCHAR( ExtractFile  ),
                                         	GENERIC_WRITE,
                                         	FILE_SHARE_WRITE,
                                         	NIL,
                                         	CREATE_ALWAYS,
                                         	FILE_ATTRIBUTE_NORMAL,
                                         	0
                                            );

                    IF OutFile <> INVALID_HANDLE_VALUE THEN
                    BEGIN

                       TRY
                          TRY

                                (* Show extracting filename in status bar *)

                             SendMessage( GetDlgItem( DlgWin, CM_STATUS  ), WM_SETTEXT, 0, LONGINT( PCHAR( 'Extracting: ' + {ExtractFilename( } ExtractFile { )}  )  )  );
                             PeekMessage( Msg, 0, 0, 0, PM_REMOVE  );
                             TranslateMessage( Msg  );
                             DispatchMessage( Msg  );

                             CRC32Val := CRC_MASK;
                             Bytes_To_Go := Header.PackedSize;		(* assign to global *)

                             IF ( Header.BitFlag AND 1  ) = 1 THEN
                                Dec( Bytes_To_Go, RAND_HEAD_LEN  );

                             CASE Header.CompressType OF
                                0: UnStore; 	(* Stored *)
                                8:	Inflate;    (* Inflate *)
                             ELSE

                                BEGIN

                                   MessageBox( 0, 'Unsupported compression method', 'Unable to extract...', mb_OK  );
                                   EXIT;

                                END;

                             END;

                                            (* Un-hilite file in listbox *)

                             IF Header.CRC32 = ( CRC32Val xor CRC_MASK )  THEN
                                SendMessage ( hWndList, LB_SETSEL, 0, Index  );

                          EXCEPT
                             //MessageBox( 0, 'Error...', 'Error', mb_OK  )
                          END;

                       FINALLY

                          IF NOT CloseHandle( OutFile ) THEN
                             MessageBox( 0, PCHAR( ExtractFile  ) {FN}, 'Error Closing...', mb_OK  );

                       END;

                    END ELSE

                       MessageBox( 0, PCHAR( ExtractFile  ) {FN}, 'Couldn''t open...', mb_OK  );

                 END ELSE
                       (* Terminate file processing if not password *)
                    BREAK;
                 //END;
              END;

           END;

   ByPass:

           FSeek( SavePos +
           			SIZEOF( Header  ) +
                    Header.FilenameLen +
                    {Header.CommentLen +}
                    Header.PackedSize,
                    FILE_BEGIN
                     );

        END;


     IF IsPassword AND NOT FillListBox THEN
			BEGIN

           IF SendMessage ( hWndList, LB_GETSELCOUNT, 0, 0  ) = 0 THEN
              //MessageBox( 0, 'CRC validation passed on file( s )', 'CRC Validation Results', mb_OK  )
           	SendMessage( GetDlgItem( DlgWin, CM_STATUS  ), WM_SETTEXT, 0, LONGINT( PCHAR( 'CRC validation passed on all file( s )... extraction successful'  )  )  )
           ELSE
              //MessageBox( 0, 'The hilited file( s ) failed CRC validation', 'CRC Validation Results', mb_OK  );
           	SendMessage( GetDlgItem( DlgWin, CM_STATUS  ), WM_SETTEXT, 0, LONGINT( PCHAR( 'The remaining hilited file( s ) failed CRC validation'  )  )  );

        END ELSE
												(* if listbox empty, exit program *)
           IF FillListBox THEN
              IF SendMessage ( hWndList, LB_GETCOUNT, 0, 0  ) = 0 THEN
              //BEGIN
                 MessageBox( 0, 'Can''t read archive header', 'Error...', mb_OK  );
              //   EndDialog( MainWin, 0  );
              //END;

  FINALLY

  	Freemem( FN );

     IF NOT HasStoredPaths THEN
     BEGIN

     								(* Set CM_STOREDPATH to Unchecked *)

     	SendMessage( GetDlgItem( DlgWin, CM_STOREDPATH  ), BM_SETCHECK, 0, 0  );

                             (* Disable the UseStoredPaths checkbox *)

  		EnableWindow( GetDlgItem( DlgWin, CM_STOREDPATH  ), FALSE  );

     END;

  END;

END (* ProcessArchive *);
(*--------------------------------------------------------------------------*)
PROCEDURE Make_CRC32Table;

VAR
	i, j: WORD;
	r: LONGINT;

CONST
	CRCPOLY    = $EDB88320;
	UCHAR_MAX  = 255;
	CHAR_BIT   = 8;

BEGIN

	FOR i := 0 to UCHAR_MAX DO
	BEGIN

		r := i;

		FOR j := CHAR_BIT DOWNTO 1 DO
     	IF (r AND 1) > 0 THEN
        	r := (r SHR 1) XOR CRCPOLY
			ELSE
         	r := r SHR 1;

		CRC32Table[i] := r;

	END;
END;
(*--------------------------------------------------------------------------*)
FUNCTION UpdC32(Octet: byte; Crc: LongInt) : LongInt;
BEGIN
	RESULT := CRC32TABLE[byte(Crc XOR LongInt(Octet))] XOR ((Crc SHR 8) AND $00FFFFFF);
END;
(*--------------------------------------------------------------------------*)
(* Update the encryption keys with the next byte of plain text *)
PROCEDURE UDK(c: byte);
BEGIN
	  key[0] := UpdC32(c, key[0]);
	  key[1] := key[1] + key[0] AND $000000ff;
	  key[1] := key[1] * 134775813 + 1;
	  //key[2] := UpdC32(key[1] shr 24, key[2]);
	  key[2] := UpdC32(HIBYTE(HIWORD(key[1])), key[2]);
END;
(*--------------------------------------------------------------------------*)
(* Initialize the encryption keys and the random header according to
	 the given password. *)
PROCEDURE seedk(passwd: STRING);

VAR
	i: BYTE;

BEGIN
	key[0] := 305419896;
	key[1] := 591751049;
	key[2] := 878082192;
	FOR i := 1 to LENGTH(passwd) do
		udk(BYTE(passwd[i]));
END;
(*--------------------------------------------------------------------------*)
(* Return the next byte in the pseudo-random sequence *)
FUNCTION decrypt_byte: INTEGER;
VAR
	temp: WORD;
BEGIN
	temp := word(key[2] or 2);
	RESULT := integer( word((temp * (temp XOR 1)) SHR 8) AND $ff);
END;
(*--------------------------------------------------------------------------*)
FUNCTION decrypt_pw(Encrypt_Head: PCHAR; EncHead_len: BYTE;
				BitFlag: WORD; CRC, FileDate: LONGINT; password: STRING): BOOLEAN;

VAR
	i,c,b: BYTE;

BEGIN

	RESULT := FALSE;
	IF password = '' THEN EXIT;

	seedk(Password);
	FOR i := 0 to EncHead_len - 1 DO
	BEGIN
		c := byte(Encrypt_Head[i + EncHead_len]) XOR decrypt_byte;
		udk(c);
		Encrypt_Head[i] := char(c);
	END;

	(* version pre 2.0 -- (currently not used) *)
	//c := byte(Encrypt_Head[EncHead_len - 2]);

	(* version 2.0+ *)
	b := BYTE(Encrypt_Head[EncHead_len - 1]);

	IF NOT ((BitFlag AND 8) = 8) THEN
	BEGIN
		//if b = (CRC shr 24) then
		IF b = hibyte(hiword(crc)) THEN
			RESULT := TRUE;

	END ELSE BEGIN
		//if c = lobyte(loword(FileDate)) then
		// b = hibyte(loword(FileDate)) then
		IF b = loword(FileDate) SHR 8 THEN
			RESULT := TRUE;

	END;

END;
(*--------------------------------------------------------------------------*)



END.

